@@ -517,6 +517,12 @@ agent_card_ecc_kem (ctrl_t ctrl, const unsigned char *ecc_ct,
 
   if (len == ecc_point_len)
     memcpy (ecc_ecdh, ecdh, len);
+  else if (len && (len - 1) * 2 == ecc_point_len - 1 && ecdh[0] == 0x02)
+    {
+      /* It's x-coordinate-only (compressed) point representation.  */
+      memcpy (ecc_ecdh, ecdh, len);
+      memset (ecc_ecdh + len, 0, ecc_point_len - len);
+    }
   else if (len == ecc_point_len + 1 && ecdh[0] == 0x40) /* The prefix */
     memcpy (ecc_ecdh, ecdh + 1, len - 1);
   else
