@@ -100,7 +100,8 @@ static unsigned int any_cert_of_class;
 
 
 #ifdef HAVE_W32_SYSTEM
-/* We load some functions dynamically.  Provide typedefs for tehse
+#include <wincrypt.h>
+/* We load some functions dynamically.  Provide typedefs for these
  * functions.  */
 typedef HCERTSTORE (WINAPI *CERTOPENSYSTEMSTORE)
   (HCRYPTPROV hProv, LPCSTR szSubsystemProtocol);
@@ -224,7 +225,7 @@ cert_compute_fpr (ksba_cert_t cert, unsigned char *digest)
 
 
 
-/* Cleanup one slot.  This releases all resourses but keeps the actual
+/* Cleanup one slot.  This releases all resources but keeps the actual
    slot in the cache marked for reuse. */
 static void
 clean_cache_slot (cert_item_t ci)
@@ -271,6 +272,21 @@ put_cert (ksba_cert_t cert, int permanent, unsigned int trustclass,
   cert_item_t ci;
   fingerprint_list_t ignored;
 
+  /* Do not keep expired certificates in the permanent cache.  */
+  if (permanent && !opt.debug_cache_expired_certs)
+    {
+      ksba_isotime_t not_after;
+      ksba_isotime_t current_time;
+
+      if (ksba_cert_get_validity (cert, 1, not_after))
+        return gpg_error (GPG_ERR_BAD_CERT);
+
+      gnupg_get_isotime (current_time);
+
+      if (*not_after && strcmp (current_time, not_after) > 0)
+        return gpg_error (GPG_ERR_CERT_EXPIRED);
+    }
+
   fpr = fpr_buffer? fpr_buffer : &help_fpr_buffer;
 
   /* If we already reached the caching limit, drop a couple of certs
@@ -753,7 +769,7 @@ cert_cache_init (strlist_t hkp_cacerts)
   initialization_done = 1;
   release_cache_lock ();
 
-  cert_cache_print_stats ();
+  cert_cache_print_stats (NULL);
 }
 
 /* Deinitialize the certificate cache.  With FULL set to true even the
@@ -796,7 +812,7 @@ cert_cache_deinit (int full)
 
 /* Print some statistics to the log file.  */
 void
-cert_cache_print_stats (void)
+cert_cache_print_stats (ctrl_t ctrl)
 {
   cert_item_t ci;
   int idx;
@@ -833,16 +849,19 @@ cert_cache_print_stats (void)
 
   release_cache_lock ();
 
-  log_info (_("permanently loaded certificates: %u\n"),
-            n_permanent);
-  log_info (_("    runtime cached certificates: %u\n"),
-            n_nonperm);
-  log_info (_("           trusted certificates: %u (%u,%u,%u,%u)\n"),
-            n_trusted,
-            n_trustclass_system,
-            n_trustclass_config,
-            n_trustclass_hkp,
-            n_trustclass_hkpspool);
+  dirmngr_status_helpf (ctrl,
+                 _("permanently loaded certificates: %u\n"),
+                        n_permanent);
+  dirmngr_status_helpf (ctrl,
+                 _("    runtime cached certificates: %u\n"),
+                        n_nonperm);
+  dirmngr_status_helpf (ctrl,
+                 _("           trusted certificates: %u (%u,%u,%u,%u)\n"),
+                        n_trusted,
+                        n_trustclass_system,
+                        n_trustclass_config,
+                        n_trustclass_hkp,
+                        n_trustclass_hkpspool);
 }
 
 
