@@ -154,6 +154,9 @@ epoch2ldaptime (time_t stamp)
 #endif
 
 
+/*
+ * Begin LDAP wrappers
+ */
 static void
 my_ldap_value_free (char **vals)
 {
@@ -162,6 +165,66 @@ my_ldap_value_free (char **vals)
 }
 
 
+/* LDAP wrappers to cope with the stupid use of ULONG instead of int in
+ * the Windows LDAP interface.  rfc1823 alsways uses int and thus
+ * ldap_parse_result should also do this.  */
+#ifdef HAVE_W32_SYSTEM
+static int
+my_ldap_return_with_check (ULONG l_err)
+{
+  if ((int)l_err < 0)
+    {
+      log_error ("oops: LDAP returned a negative error code (0x%lx)\n", l_err);
+      l_err = LDAP_OTHER;
+    }
+  return (int)l_err;
+}
+#endif
+
+static int
+my_ldap_parse_result (LDAP *ld, LDAPMessage *result,
+                      int *errcodep, char **matcheddnp, char **errmsgp,
+                      char ***referralsp, LDAPControl ***serverctrlsp,
+                      int freeit)
+{
+#ifdef HAVE_W32_SYSTEM
+  ULONG l_err;
+  ULONG l_errcode;
+  l_err = ldap_parse_result (ld, result,
+                             errcodep? &l_errcode : NULL,
+                             matcheddnp, errmsgp,
+                             referralsp, serverctrlsp, freeit);
+  if (errcodep)
+    *errcodep = l_errcode;
+  return my_ldap_return_with_check (l_err);
+#else
+  return ldap_parse_result (ld, result, errcodep, matcheddnp, errmsgp,
+                            referralsp, serverctrlsp, freeit);
+#endif
+}
+
+
+static int
+my_ldap_parse_page_control (LDAP *ld, LDAPControl **ctrls,
+                            int *count, struct berval **cookie)
+{
+#ifdef HAVE_W32_SYSTEM
+  ULONG l_err;
+  ULONG l_count;
+  l_err = ldap_parse_page_control (ld, ctrls, count? &l_count: NULL, cookie);
+  if (count)
+    *count = l_count;
+  return my_ldap_return_with_check (l_err);
+#else
+  return ldap_parse_page_control (ld, ctrls, count, cookie);
+#endif
+}
+
+/*
+ * End LDAP wrappers
+ */
+
+
 /* Print a description of supported variables.  */
 void
 ks_ldap_help_variables (ctrl_t ctrl)
@@ -1356,10 +1419,11 @@ search_and_parse (ctrl_t ctrl, const char *keyspec,
                   char **attrs, LDAPMessage **r_message)
 {
   gpg_error_t err = 0;
-  int l_err, l_reserr;
+  int l_err;
+  int l_reserr;
+  unsigned int totalcount = 0;
   LDAPControl *srvctrls[2] = { NULL, NULL };
   int count;
-  unsigned int totalcount = 0;
   LDAPControl *pagectrl = NULL;
   LDAPControl **resctrls = NULL;
 
@@ -1399,8 +1463,8 @@ search_and_parse (ctrl_t ctrl, const char *keyspec,
 
   if (ctrl->ks_get_state)
     {
-      l_err = ldap_parse_result (ldap_conn, *r_message, &l_reserr,
-                                 NULL, NULL, NULL, &resctrls, 0);
+      l_err = my_ldap_parse_result (ldap_conn, *r_message, &l_reserr,
+                                    NULL, NULL, NULL, &resctrls, 0);
       if (l_err)
         {
           err = ldap_err_to_gpg_err (l_err);
@@ -1414,9 +1478,9 @@ search_and_parse (ctrl_t ctrl, const char *keyspec,
           ber_bvfree (ctrl->ks_get_state->pagecookie);
           ctrl->ks_get_state->pagecookie = NULL;
         }
-      l_err = ldap_parse_page_control (ldap_conn, resctrls,
-                                       &totalcount,
-                                       &ctrl->ks_get_state->pagecookie);
+      l_err = my_ldap_parse_page_control (ldap_conn, resctrls,
+                                          &totalcount,
+                                          &ctrl->ks_get_state->pagecookie);
       if (l_err)
         {
           err = ldap_err_to_gpg_err (l_err);
@@ -2984,6 +3048,116 @@ ks_ldap_put (ctrl_t ctrl, parsed_uri_t uri,
 }
 
 
+/* Delete the keys given by the list of fingerprints in FPRLIST from
+ * the keyserver identified by URI.  The function stops at the first
+ * error encountered. */
+gpg_error_t
+ks_ldap_del (ctrl_t ctrl, parsed_uri_t uri, strlist_t fprlist)
+{
+  gpg_error_t err = 0;
+  int ldap_err;
+  unsigned int serverinfo;
+  LDAP *ldap_conn = NULL;
+  char *basedn = NULL;
+  char *dn = NULL;
+  strlist_t fpr;
+  unsigned int count = 0;
+  unsigned int totalcount = 0;
+
+  if (dirmngr_use_tor ())
+    {
+      return no_ldap_due_to_tor (ctrl);
+    }
+
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    totalcount++;
+
+  err = my_ldap_connect (uri, 0, &ldap_conn, &basedn, NULL, NULL, &serverinfo);
+  if (err || !basedn)
+    {
+      if(opt.verbose)
+        log_info ("%s: connecting to server failed\n", __func__);
+      if (!err)
+	err = gpg_error (GPG_ERR_GENERAL);  /* (no baseDN) */
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_REALLDAP))
+    {
+      if(opt.verbose)
+        log_info ("%s: The PGP.com keyserver is not supported\n", __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (!(serverinfo & SERVERINFO_SCHEMAV2))
+    {
+      if(opt.verbose)
+        log_info ("%s: The keyserver does not support the v2 schema\n",
+                  __func__);
+      err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      goto leave;
+    }
+
+  if (opt.verbose)
+    log_info ("%s: Using DN: %s,%s\n", __func__,
+              (serverinfo & SERVERINFO_NTDS)? "CN=<fingerprint>"
+              /* */                         : "pgpCertID=<keyid>",
+              basedn);
+  for (fpr = fprlist; fpr; fpr = fpr->next)
+    {
+      if ((serverinfo & SERVERINFO_NTDS))
+        {
+          xfree (dn);
+          dn = xtryasprintf ("CN=%s,%s", fpr->d, basedn);
+        }
+      else
+        {
+          unsigned int off;
+
+          /* Simle method to get the keyID.  Note that a v5 key
+           * (len>40) has the keyid at the left.  If the length is
+           * less than 17 we assume a keyid has been given.  */
+          off = strlen (fpr->d);
+          if (off <= 40 && off > 16)
+            off = off - 16;
+          else
+            off = 0;
+
+          xfree (dn);
+          dn = xtryasprintf ("pgpCertID=%.16s,%s", fpr->d+off, basedn);
+        }
+
+      npth_unprotect ();
+      ldap_err = ldap_delete_ext_s (ldap_conn, dn, NULL, NULL);
+      npth_protect ();
+      if (ldap_err == LDAP_SUCCESS)
+        {
+          if (opt.verbose)
+            log_info ("%s: key %s deleted\n", __func__, fpr->d);
+          count++;
+        }
+      else
+        {
+          log_error ("%s: error deleting key %s: %s\n",
+                     __func__, fpr->d, ldap_err2string (ldap_err));
+          err = ldap_err_to_gpg_err (ldap_err);
+          break;  /* Stop at the first failed deletion.  */
+        }
+    }
+  log_info ("%s: number of keys deleted: %u of %u\n",
+            __func__, count, totalcount);
+
+
+ leave:
+  if (ldap_conn)
+    ldap_unbind (ldap_conn);
+  xfree (dn);
+  xfree (basedn);
+  return err;
+}
+
+
 
 /* Get the data described by FILTER_ARG from URI.  On success R_FP has
  * an open stream to read the data.  KS_GET_FLAGS conveys flags from
