@@ -214,6 +214,7 @@ struct
   int no_encrypt_to;
   int encrypt_to_default_key;
   int interactive;
+  strlist_t print_notations;       /* Name of notations to print. */
   struct notation *sig_notations;
   struct notation *cert_notations;
   strlist_t sig_policy_url;
@@ -368,6 +369,7 @@ struct {
 #define DBG_TRUST_VALUE   256	/* debug the trustdb */
 #define DBG_HASHING_VALUE 512	/* debug hashing operations */
 #define DBG_IPC_VALUE     1024  /* debug assuan communication */
+#define DBG_RECSEL_VALUE  2048  /* Debug the record selection */
 #define DBG_CLOCK_VALUE   4096
 #define DBG_LOOKUP_VALUE  8192	/* debug the key lookup */
 #define DBG_EXTPROG_VALUE 16384 /* debug external program calls */
@@ -382,6 +384,7 @@ struct {
 #define DBG_TRUST  (opt.debug & DBG_TRUST_VALUE)
 #define DBG_HASHING (opt.debug & DBG_HASHING_VALUE)
 #define DBG_IPC     (opt.debug & DBG_IPC_VALUE)
+#define DBG_RECSEL  (opt.debug & DBG_RECSEL_VALUE)
 #define DBG_CLOCK   (opt.debug & DBG_CLOCK_VALUE)
 #define DBG_LOOKUP  (opt.debug & DBG_LOOKUP_VALUE)
 #define DBG_EXTPROG (opt.debug & DBG_EXTPROG_VALUE)
@@ -397,7 +400,8 @@ EXTERN_UNLESS_MAIN_MODULE int memory_stat_debug_mode;
 /* Compatibility flags */
 #define COMPAT_PARALLELIZED   1  /* Use threaded hashing for signatures.  */
 #define COMPAT_T7014_OLD      2  /* Use initial T7014 test data.  */
-
+#define COMPAT_COMPR_KEYS     4  /* Allow import of compressed keys. (T7014) */
+#define COMPAT_NO_MANU        8  /* Do not include a "manu" notation.  */
 
 /* Compliance test macros.  */
 #define GNUPG   (opt.compliance==CO_GNUPG || opt.compliance==CO_DE_VS)
@@ -463,6 +467,8 @@ EXTERN_UNLESS_MAIN_MODULE int memory_stat_debug_mode;
 #define LIST_SHOW_X509_NOTATIONS         (1<<17)
 #define LIST_STORE_X509_NOTATIONS        (1<<18)
 #define LIST_SHOW_OWNERTRUST             (1<<19)
+#define LIST_SHOW_TRUSTSIG               (1<<20)
+#define LIST_SHOW_HIDDEN_NOTATIONS       (1<<21)
 
 #define VERIFY_SHOW_PHOTOS               (1<<0)
 #define VERIFY_SHOW_POLICY_URLS          (1<<1)
@@ -473,6 +479,7 @@ EXTERN_UNLESS_MAIN_MODULE int memory_stat_debug_mode;
 #define VERIFY_SHOW_UID_VALIDITY         (1<<5)
 #define VERIFY_SHOW_UNUSABLE_UIDS        (1<<6)
 #define VERIFY_SHOW_PRIMARY_UID_ONLY     (1<<9)
+#define VERIFY_SHOW_HIDDEN_NOTATIONS     (1<<21)
 
 #define KEYSERVER_HTTP_PROXY             (1<<0)
 #define KEYSERVER_TIMEOUT                (1<<1)
