@@ -33,11 +33,6 @@
 #include "options.h"
 
 
-/* Maximum buffer sizes required for ECC KEM.  */
-#define ECC_POINT_LEN_MAX (1+2*64)
-#define ECC_HASH_LEN_MAX 64
-
-
 /* FIXME: Better change the function name because mpi_ is used by
    gcrypt macros.  */
 gcry_mpi_t
@@ -183,29 +178,6 @@ sexp_extract_param_sos_nlz (gcry_sexp_t sexp, const char *param,
 }
 
 
-static byte *
-get_data_from_sexp (gcry_sexp_t sexp, const char *item, size_t *r_size)
-{
-  gcry_sexp_t list;
-  size_t valuelen;
-  const char *value;
-  byte *v;
-
-  if (DBG_CRYPTO)
-    log_printsexp ("get_data_from_sexp:", sexp);
-
-  list = gcry_sexp_find_token (sexp, item, 0);
-  log_assert (list);
-  value = gcry_sexp_nth_data (list, 1, &valuelen);
-  log_assert (value);
-  v = xtrymalloc (valuelen);
-  memcpy (v, value, valuelen);
-  gcry_sexp_release (list);
-  *r_size = valuelen;
-  return v;
-}
-
-
 /****************
  * Emulate our old PK interface here - sometime in the future we might
  * change the internal design to directly fit to libgcrypt.
@@ -331,9 +303,18 @@ pk_verify (pubkey_algo_t pkalgo, gcry_mpi_t hash,
             rc = gpg_error (GPG_ERR_BAD_MPI);
           else
             {
+              r = gcry_mpi_copy (r);
+              s = gcry_mpi_copy (s);
+
+              if (!r || !s)
+                {
+                  rc = gpg_error_from_syserror ();
+                  goto leave;
+                }
+
               /* We need to fixup the length in case of leading zeroes.
                * OpenPGP does not allow leading zeroes and the parser for
-               * the signature packet has no information on the use curve,
+               * the signature packet has no information on the used curve,
                * thus we need to do it here.  We won't do it for opaque
                * MPIs under the assumption that they are known to be fine;
                * we won't see them here anyway but the check is anyway
@@ -343,12 +324,13 @@ pk_verify (pubkey_algo_t pkalgo, gcry_mpi_t hash,
 
               if (rlen < neededfixedlen
                   && !gcry_mpi_get_flag (r, GCRYMPI_FLAG_OPAQUE)
-                  && !(rc=gcry_mpi_print (GCRYMPI_FMT_USG, buf, sizeof buf, &n, r)))
+                  && !(rc=gcry_mpi_print (GCRYMPI_FMT_USG,
+                                          buf, sizeof buf, &n, r)))
                 {
                   log_assert (n < neededfixedlen);
                   memmove (buf + (neededfixedlen - n), buf, n);
                   memset (buf, 0, neededfixedlen - n);
-                  r = gcry_mpi_set_opaque_copy (NULL, buf, neededfixedlen * 8);
+                  gcry_mpi_set_opaque_copy (r, buf, neededfixedlen * 8);
                 }
               else if (rlen < neededfixedlen
                        && gcry_mpi_get_flag (r, GCRYMPI_FLAG_OPAQUE))
@@ -361,14 +343,18 @@ pk_verify (pubkey_algo_t pkalgo, gcry_mpi_t hash,
                   memset (buf, 0, neededfixedlen - n);
                   gcry_mpi_set_opaque_copy (r, buf, neededfixedlen * 8);
                 }
-              if (slen < neededfixedlen
+
+              if (rc)
+                ;
+              else if (slen < neededfixedlen
                   && !gcry_mpi_get_flag (s, GCRYMPI_FLAG_OPAQUE)
-                  && !(rc=gcry_mpi_print (GCRYMPI_FMT_USG, buf, sizeof buf, &n, s)))
+                  && !(rc=gcry_mpi_print (GCRYMPI_FMT_USG,
+                                          buf, sizeof buf, &n, s)))
                 {
                   log_assert (n < neededfixedlen);
                   memmove (buf + (neededfixedlen - n), buf, n);
                   memset (buf, 0, neededfixedlen - n);
-                  s = gcry_mpi_set_opaque_copy (NULL, buf, neededfixedlen * 8);
+                  gcry_mpi_set_opaque_copy (s, buf, neededfixedlen * 8);
                 }
               else if (slen < neededfixedlen
                        && gcry_mpi_get_flag (s, GCRYMPI_FLAG_OPAQUE))
@@ -416,6 +402,7 @@ pk_verify (pubkey_algo_t pkalgo, gcry_mpi_t hash,
   if (!rc)
     rc = gcry_pk_verify (s_sig, s_hash, s_pkey);
 
+ leave:
   gcry_sexp_release (s_sig);
   gcry_sexp_release (s_hash);
   gcry_sexp_release (s_pkey);
@@ -440,7 +427,9 @@ do_encrypt_kem (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
   gcry_sexp_t s_data = NULL;
   gcry_cipher_hd_t hd = NULL;
   char *ecc_oid = NULL;
-  enum gcry_kem_algos kyber_algo, ecc_algo;
+  const char *curve;
+  const struct gnupg_ecc_params *ecc;
+  enum gcry_kem_algos kyber_algo;
 
   const unsigned char *ecc_pubkey;
   size_t ecc_pubkey_len;
@@ -481,103 +470,56 @@ do_encrypt_kem (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
       log_error ("%s: error getting OID for ECC key\n", __func__);
       goto leave;
     }
-  ecc_algo = openpgp_oid_to_kem_algo (ecc_oid);
-  if (ecc_algo == GCRY_KEM_RAW_X25519)
+  curve = openpgp_oid_to_curve (ecc_oid, 1);
+  if (!curve)
+    {
+      err = gpg_error (GPG_ERR_INV_DATA);
+      log_error ("%s: error getting curve for ECC key\n", __func__);
+      goto leave;
+    }
+  ecc = gnupg_get_ecc_params (curve);
+  if (!ecc)
+    {
+      if (opt.verbose)
+        log_info ("%s: ECC curve %s not supported\n", __func__, curve);
+      err = gpg_error (GPG_ERR_INV_DATA);
+      goto leave;
+    }
+  ecc_ct_len = ecc_ecdh_len = ecc->point_len;
+  ecc_ss_len = ecc->scalar_len;
+  ecc_hash_algo = ecc->hash_algo;
+
+  ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
+  ecc_pubkey_len = (nbits+7)/8;
+  if (ecc_pubkey_len != ecc->pubkey_len)
+    {
+      if (ecc->kem_algo == GCRY_KEM_RAW_X25519
+          && ecc_pubkey_len == ecc->pubkey_len - 1)
+        /* For Curve25519, we also accept no prefix in the point
+         * representation.  */
+        ;
+      else
+        {
+          if (opt.verbose)
+            log_info ("%s: ECC public key length invalid (%zu)\n",
+                      __func__, ecc_pubkey_len);
+          err = gpg_error (GPG_ERR_INV_DATA);
+          goto leave;
+        }
+    }
+
+  if (ecc->kem_algo == GCRY_KEM_RAW_X25519)
     {
       if (!strcmp (ecc_oid, "1.3.6.1.4.1.3029.1.5.1"))
         log_info ("Warning: "
                   "legacy OID for cv25519 accepted during development\n");
-      ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
-      ecc_pubkey_len = (nbits+7)/8;
+      /* Optional prefix handling */
       if (ecc_pubkey_len == 33 && *ecc_pubkey == 0x40)
         {
           ecc_pubkey++;     /* Remove the 0x40 prefix.  */
           ecc_pubkey_len--;
         }
-      if (ecc_pubkey_len != 32)
-        {
-          if (opt.verbose)
-            log_info ("%s: ECC public key length invalid (%zu)\n",
-                      __func__, ecc_pubkey_len);
-          err = gpg_error (GPG_ERR_INV_DATA);
-          goto leave;
-        }
-      ecc_ct_len = ecc_ecdh_len = 32;
-      ecc_ss_len = 32;
-      ecc_hash_algo = GCRY_MD_SHA3_256;
     }
-  else if (ecc_algo == GCRY_KEM_RAW_X448)
-    {
-      ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
-      ecc_pubkey_len = (nbits+7)/8;
-      if (ecc_pubkey_len != 56)
-        {
-          if (opt.verbose)
-            log_info ("%s: ECC public key length invalid (%zu)\n",
-                      __func__, ecc_pubkey_len);
-          err = gpg_error (GPG_ERR_INV_DATA);
-          goto leave;
-        }
-      ecc_ct_len = ecc_ecdh_len = 56;
-      ecc_ss_len = 64;
-      ecc_hash_algo = GCRY_MD_SHA3_512;
-    }
-  else if (ecc_algo == GCRY_KEM_RAW_BP256)
-    {
-      ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
-      ecc_pubkey_len = (nbits+7)/8;
-      if (ecc_pubkey_len != 65)
-        {
-          if (opt.verbose)
-            log_info ("%s: ECC public key length invalid (%zu)\n",
-                      __func__, ecc_pubkey_len);
-          err = gpg_error (GPG_ERR_INV_DATA);
-          goto leave;
-        }
-      ecc_ct_len = ecc_ecdh_len = 65;
-      ecc_ss_len = 32;
-      ecc_hash_algo = GCRY_MD_SHA3_256;
-    }
-  else if (ecc_algo == GCRY_KEM_RAW_BP384)
-    {
-      ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
-      ecc_pubkey_len = (nbits+7)/8;
-      if (ecc_pubkey_len != 97)
-        {
-          if (opt.verbose)
-            log_info ("%s: ECC public key length invalid (%zu)\n",
-                      __func__, ecc_pubkey_len);
-          err = gpg_error (GPG_ERR_INV_DATA);
-          goto leave;
-        }
-      ecc_ct_len = ecc_ecdh_len = 97;
-      ecc_ss_len = 64;
-      ecc_hash_algo = GCRY_MD_SHA3_512;
-    }
-  else if (ecc_algo == GCRY_KEM_RAW_BP512)
-    {
-      ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
-      ecc_pubkey_len = (nbits+7)/8;
-      if (ecc_pubkey_len != 129)
-        {
-          if (opt.verbose)
-            log_info ("%s: ECC public key length invalid (%zu)\n",
-                      __func__, ecc_pubkey_len);
-          err = gpg_error (GPG_ERR_INV_DATA);
-          goto leave;
-        }
-      ecc_ct_len = ecc_ecdh_len = 129;
-      ecc_ss_len = 64;
-      ecc_hash_algo = GCRY_MD_SHA3_512;
-    }
-  else
-    {
-      if (opt.verbose)
-        log_info ("%s: ECC curve %s not supported\n", __func__, ecc_oid);
-      err = gpg_error (GPG_ERR_INV_DATA);
-      goto leave;
-    }
-
 
   if (DBG_CRYPTO)
     {
@@ -585,7 +527,7 @@ do_encrypt_kem (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
       log_printhex (ecc_pubkey, ecc_pubkey_len, "ECC   pubkey:");
     }
 
-  err = gcry_kem_encap (ecc_algo,
+  err = gcry_kem_encap (ecc->kem_algo,
                         ecc_pubkey, ecc_pubkey_len,
                         ecc_ct, ecc_ct_len,
                         ecc_ecdh, ecc_ecdh_len,
@@ -606,7 +548,7 @@ do_encrypt_kem (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
                            ecc_hash_algo,
                            ecc_ecdh, ecc_ecdh_len,
                            ecc_ct, ecc_ct_len,
-                           ecc_pubkey, ecc_pubkey_len);
+                           ecc_pubkey, ecc_pubkey_len, NULL, 0);
   if (err)
     {
       if (opt.verbose)
@@ -758,95 +700,230 @@ do_encrypt_kem (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
 static gpg_error_t
 do_encrypt_ecdh (PKT_public_key *pk, gcry_mpi_t data,  gcry_mpi_t *resarr)
 {
-  gcry_mpi_t *pkey   = pk->pkey;
-  gcry_sexp_t s_ciph = NULL;
-  gcry_sexp_t s_data = NULL;
-  gcry_sexp_t s_pkey = NULL;
   gpg_error_t err;
-  gcry_mpi_t k = NULL;
-  char *curve = NULL;
-  int with_djb_tweak_flag;
-  gcry_mpi_t public = NULL;
-  gcry_mpi_t result = NULL;
-  byte fp[MAX_FINGERPRINT_LEN];
-  byte *shared = NULL;
-  byte *p;
-  size_t nshared;
   unsigned int nbits;
+  gcry_cipher_hd_t hd = NULL;
+  char *ecc_oid = NULL;
+  const char *curve;
+  const struct gnupg_ecc_params *ecc;
 
-  err = pk_ecdh_generate_ephemeral_key (pkey, &k);
-  if (err)
-    goto leave;
+  const unsigned char *ecc_pubkey;
+  size_t ecc_pubkey_len;
+  const unsigned char *seskey;
+  size_t seskey_len;
+  unsigned char *enc_seskey = NULL;
+  size_t enc_seskey_len;
 
-  curve = openpgp_oid_to_str (pkey[0]);
-  if (!curve)
-    {
-      err = gpg_error_from_syserror ();
-      goto leave;
-    }
+  unsigned char ecc_ct[ECC_POINT_LEN_MAX];
+  unsigned char ecc_ecdh[ECC_POINT_LEN_MAX];
+  size_t ecc_ct_len, ecc_ecdh_len;
 
-  with_djb_tweak_flag = openpgp_oid_is_cv25519 (pkey[0]);
+  unsigned char *kek = NULL;
+  size_t kek_len;
 
-  /* Now use the ephemeral secret to compute the shared point.  */
-  err = gcry_sexp_build (&s_pkey, NULL,
-                         with_djb_tweak_flag ?
-                         "(public-key(ecdh(curve%s)(flags djb-tweak)(q%m)))"
-                         : "(public-key(ecdh(curve%s)(q%m)))",
-                         curve, pkey[1]);
-  if (err)
-    goto leave;
-
-  /* Put K into a simplified S-expression.  */
-  err = gcry_sexp_build (&s_data, NULL, "%m", k);
-  if (err)
-    goto leave;
-
-  /* Run encryption.  */
-  err = gcry_pk_encrypt (&s_ciph, s_data, s_pkey);
-  if (err)
-    goto leave;
-
-  gcry_sexp_release (s_data); s_data = NULL;
-  gcry_sexp_release (s_pkey); s_pkey = NULL;
-
-
-  /* Get the shared point and the ephemeral public key.  */
-  shared = get_data_from_sexp (s_ciph, "s", &nshared);
-  if (!shared)
-    {
-      err = gpg_error_from_syserror ();
-      goto leave;
-    }
-  err = sexp_extract_param_sos (s_ciph, "e", &public);
-  gcry_sexp_release (s_ciph); s_ciph = NULL;
-  if (DBG_CRYPTO)
-    {
-      log_debug ("ECDH ephemeral key:");
-      gcry_mpi_dump (public);
-      log_printf ("\n");
-    }
+  const unsigned char *kdf_params_spec;
+  byte fp[MAX_FINGERPRINT_LEN];
+  int keywrap_cipher_algo;
+  int kdf_hash_algo;
+  unsigned char *kdf_params = NULL;
+  size_t kdf_params_len = 0;
 
   fingerprint_from_pk (pk, fp, NULL);
 
-  p = gcry_mpi_get_opaque (data, &nbits);
-  result = NULL;
-  err = pk_ecdh_encrypt_with_shared_point (shared, nshared, fp, p,
-                                           (nbits+7)/8, pkey, &result);
-  if (err)
-    goto leave;
+  ecc_oid = openpgp_oid_to_str (pk->pkey[0]);
+  if (!ecc_oid)
+    {
+      err = gpg_error_from_syserror ();
+      log_error ("%s: error getting OID for ECC key\n", __func__);
+      goto leave;
+    }
+  curve = openpgp_oid_to_curve (ecc_oid, 1);
+  if (!curve)
+    {
+      err = gpg_error (GPG_ERR_INV_DATA);
+      log_error ("%s: error getting curve for ECC key\n", __func__);
+      goto leave;
+    }
+  ecc = gnupg_get_ecc_params (curve);
+  if (!ecc)
+    {
+      if (opt.verbose)
+        log_info ("%s: ECC curve %s not supported\n", __func__, curve);
+      err = gpg_error (GPG_ERR_INV_DATA);
+      goto leave;
+    }
+  ecc_ct_len = ecc_ecdh_len = ecc->point_len;
 
-  resarr[0] = public; public = NULL;
-  resarr[1] = result; result = NULL;
+  ecc_pubkey = gcry_mpi_get_opaque (pk->pkey[1], &nbits);
+  ecc_pubkey_len = (nbits+7)/8;
+  if (ecc_pubkey_len != ecc->pubkey_len)
+    {
+      if (ecc->kem_algo == GCRY_KEM_RAW_X25519
+          && ecc_pubkey_len == ecc->pubkey_len - 1)
+        /* For Curve25519, we also accept no prefix in the point
+         * representation.  */
+        ;
+      else
+        {
+          if (opt.verbose)
+            log_info ("%s: ECC public key length invalid (%zu)\n",
+                      __func__, ecc_pubkey_len);
+          err = gpg_error (GPG_ERR_INV_DATA);
+          goto leave;
+        }
+    }
+
+  if (ecc->kem_algo == GCRY_KEM_RAW_X25519)
+    {
+      /* Note: Legacy OID is OK here.  */
+      /* Optional prefix handling */
+      if (ecc_pubkey_len == 33 && *ecc_pubkey == 0x40)
+        {
+          ecc_pubkey++;     /* Remove the 0x40 prefix.  */
+          ecc_pubkey_len--;
+        }
+    }
+
+  if (DBG_CRYPTO)
+    {
+      log_debug ("ECC    curve: %s\n", ecc_oid);
+      log_printhex (ecc_pubkey, ecc_pubkey_len, "ECC   pubkey:");
+    }
+
+  err = gcry_kem_encap (ecc->kem_algo,
+                        ecc_pubkey, ecc_pubkey_len,
+                        ecc_ct, ecc_ct_len,
+                        ecc_ecdh, ecc_ecdh_len,
+                        NULL, 0);
+  if (err)
+    {
+      if (opt.verbose)
+        log_info ("%s: gcry_kem_encap for ECC (%s) failed\n",
+                  __func__, ecc_oid);
+      goto leave;
+    }
+  if (DBG_CRYPTO)
+    {
+      log_printhex (ecc_ct, ecc_ct_len, "ECC    ephem:");
+      log_printhex (ecc_ecdh, ecc_ecdh_len, "ECC     ecdh:");
+    }
+
+  err = ecc_build_kdf_params (&kdf_params, &kdf_params_len,
+                              &kdf_params_spec, pk->pkey, fp);
+  if (err)
+    return err;
+
+  keywrap_cipher_algo = kdf_params_spec[3];
+  kdf_hash_algo = kdf_params_spec[2];
+
+  if (DBG_CRYPTO)
+    log_debug ("ecdh KDF algorithms %s+%s with aeswrap\n",
+               openpgp_md_algo_name (kdf_hash_algo),
+               openpgp_cipher_algo_name (keywrap_cipher_algo));
+
+  if (kdf_hash_algo != GCRY_MD_SHA256
+      && kdf_hash_algo != GCRY_MD_SHA384
+      && kdf_hash_algo != GCRY_MD_SHA512)
+    {
+      err = gpg_error (GPG_ERR_BAD_PUBKEY);
+      goto leave;
+    }
+
+  if (keywrap_cipher_algo != CIPHER_ALGO_AES
+      && keywrap_cipher_algo != CIPHER_ALGO_AES192
+      && keywrap_cipher_algo != CIPHER_ALGO_AES256)
+    {
+      err = gpg_error (GPG_ERR_BAD_PUBKEY);
+      goto leave;
+    }
+
+  kek_len = gcry_cipher_get_algo_keylen (keywrap_cipher_algo);
+  if (kek_len > gcry_md_get_algo_dlen (kdf_hash_algo))
+    {
+      err = gpg_error (GPG_ERR_BAD_PUBKEY);
+      goto leave;
+    }
+
+  kek = xtrymalloc (kek_len);
+  if (!kek)
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+
+  err = gnupg_ecc_kem_kdf (kek, kek_len, kdf_hash_algo,
+                           ecc->is_weierstrauss ?
+                           ecc_ecdh + 1 : ecc_ecdh,
+                           ecc->is_weierstrauss ?
+                           (ecc_ecdh_len - 1) / 2 : ecc_ecdh_len,
+                           NULL, 0, NULL, 0,
+                           kdf_params, kdf_params_len);
+  xfree (kdf_params);
+  if (err)
+    {
+      if (opt.verbose)
+        log_info ("%s: kdf for ECC failed\n", __func__);
+      goto leave;
+    }
+
+  if (DBG_CRYPTO)
+    log_printhex (kek, kek_len, "KEK:");
+
+  err = gcry_cipher_open (&hd, keywrap_cipher_algo,
+                          GCRY_CIPHER_MODE_AESWRAP, 0);
+  if (!err)
+    err = gcry_cipher_setkey (hd, kek, kek_len);
+  if (err)
+    {
+      if (opt.verbose)
+        log_error ("%s: failed to initialize AESWRAP: %s\n", __func__,
+                   gpg_strerror (err));
+      goto leave;
+    }
+
+  seskey = gcry_mpi_get_opaque (data, &nbits);
+  seskey_len = (nbits+7)/8;
+
+  enc_seskey_len = 1 + seskey_len + 8;
+  enc_seskey = xtrymalloc (enc_seskey_len);
+  if (!enc_seskey || enc_seskey_len > 254)
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+
+  enc_seskey[0] = enc_seskey_len - 1;
+  err = gcry_cipher_encrypt (hd, enc_seskey+1, enc_seskey_len-1,
+                             seskey, seskey_len);
+  if (err)
+    {
+      log_error ("%s: wrapping session key failed\n", __func__);
+      goto leave;
+    }
+  if (DBG_CRYPTO)
+    log_printhex (enc_seskey, enc_seskey_len, "enc_seskey:");
+
+  resarr[0] = gcry_mpi_set_opaque_copy (NULL, ecc_ct, 8 * ecc_ct_len);
+  if (!resarr[0])
+    {
+      err = gpg_error_from_syserror ();
+      goto leave;
+    }
+
+  resarr[1] = gcry_mpi_set_opaque_copy (NULL, enc_seskey, 8 * enc_seskey_len);
+  if (!resarr[1])
+    {
+      err = gpg_error_from_syserror ();
+      gcry_mpi_release (resarr[0]);
+    }
 
  leave:
-  gcry_mpi_release (public);
-  gcry_mpi_release (result);
-  xfree (shared);
-  gcry_sexp_release (s_ciph);
-  gcry_sexp_release (s_data);
-  gcry_sexp_release (s_pkey);
-  xfree (curve);
-  gcry_mpi_release (k);
+  xfree (enc_seskey);
+  gcry_cipher_close (hd);
+  xfree (kek);
+  wipememory (ecc_ct, sizeof ecc_ct);
+  wipememory (ecc_ecdh, sizeof ecc_ecdh);
+  xfree (ecc_oid);
   return err;
 }
 
@@ -902,9 +979,9 @@ do_encrypt_rsa_elg (PKT_public_key *pk, gcry_mpi_t data, gcry_mpi_t *resarr)
  * change the internal design to directly fit to libgcrypt.  PK is is
  * the OpenPGP public key packet, DATA is an MPI with the to be
  * encrypted data, and RESARR receives the encrypted data.  RESARRAY
- * is expected to be an two item array which will be filled with newly
- * allocated MPIs.  SESKEY_ALGO is required for public key algorithms
- * which do not encode it in DATA.
+ * is expected to be an two/three item array which will be filled with
+ * newly allocated MPIs.  SESKEY_ALGO is required for public key
+ * algorithms which do not encode it in DATA.
  */
 gpg_error_t
 pk_encrypt (PKT_public_key *pk, gcry_mpi_t data, int seskey_algo,
