@@ -77,6 +77,8 @@ enum cmd_and_opt_values
     oBlacklist,
     oNoAutostart,
     oAddRevocs,
+    oNoAddRevocs,
+    oRealClean,
 
     oDummy
   };
@@ -120,7 +122,9 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_n (oWithColons, "with-colons", "@"),
   ARGPARSE_s_s (oBlacklist, "blacklist", "@"),
   ARGPARSE_s_s (oDirectory, "directory", "@"),
-  ARGPARSE_s_n (oAddRevocs, "add-revocs", "add revocation certificates"),
+  ARGPARSE_s_n (oAddRevocs, "add-revocs", "@"),
+  ARGPARSE_s_n (oNoAddRevocs, "no-add-revocs", "do not add revocation certificates"),
+  ARGPARSE_s_n (oRealClean, "realclean", "remove most key signatures"),
 
   ARGPARSE_s_s (oFakeSubmissionAddr, "fake-submission-addr", "@"),
 
@@ -151,20 +155,24 @@ const char *fake_submission_addr;
 static char **blacklist_array;
 static size_t blacklist_array_len;
 
+/* The current locale in the short form (e.g. "de" instead of "de_DE")  */
+static char *short_locale;
 
-static void wrong_args (const char *text) GPGRT_ATTR_NORETURN;
+static void wrong_args (const char *t1, const char *t2) GPGRT_ATTR_NORETURN;
 static void add_blacklist (const char *fname);
 static gpg_error_t proc_userid_from_stdin (gpg_error_t (*func)(const char *),
                                            const char *text);
 static gpg_error_t command_supported (char *userid);
 static gpg_error_t command_check (char *userid);
-static gpg_error_t command_send (const char *fingerprint, const char *userid);
+static gpg_error_t command_create (const char *fingerprint, const char *userid);
 static gpg_error_t encrypt_response (estream_t *r_output, estream_t input,
                                      const char *addrspec,
                                      const char *fingerprint);
 static gpg_error_t read_confirmation_request (estream_t msg);
 static gpg_error_t command_receive_cb (void *opaque,
-                                       const char *mediatype, estream_t fp,
+                                       const char *mediatype,
+                                       const char *language,
+                                       estream_t fp,
                                        unsigned int flags);
 static gpg_error_t command_mirror (char *domain[]);
 
@@ -202,10 +210,15 @@ my_strusage( int level )
 
 
 static void
-wrong_args (const char *text)
+wrong_args (const char *text, const char *text2)
 {
-  es_fprintf (es_stderr, _("usage: %s [options] %s\n"),
-              gpgrt_strusage (11), text);
+#if GPGRT_VERSION_NUMBER >= 0x013000 /* >= 1.48 */
+  /* Skip the leading dashes if build with command support.  */
+  if (text[0] == '-' && text[1] == '-' && text[2])
+    text += 2;
+#endif
+  es_fprintf (es_stderr, _("usage: %s %s [options] %s\n"),
+              gpgrt_strusage (11), text, text2);
   exit (2);
 }
 
@@ -233,16 +246,16 @@ parse_arguments (gpgrt_argparse_t *pargs, gpgrt_opt_t *popts)
           break;
 
         case oGpgProgram:
-          opt.gpg_program = pargs->r.ret_str;
+          opt.gpg_program = make_filename (pargs->r.ret_str, NULL);
           break;
         case oDirectory:
-          opt.directory = pargs->r.ret_str;
+          opt.directory = make_filename (pargs->r.ret_str, NULL);
           break;
         case oSend:
           opt.use_sendmail = 1;
           break;
         case oOutput:
-          opt.output = pargs->r.ret_str;
+          opt.output = make_filename (pargs->r.ret_str, NULL);
           break;
         case oFakeSubmissionAddr:
           fake_submission_addr = pargs->r.ret_str;
@@ -262,6 +275,13 @@ parse_arguments (gpgrt_argparse_t *pargs, gpgrt_opt_t *popts)
         case oAddRevocs:
           opt.add_revocs = 1;
           break;
+        case oNoAddRevocs:
+          opt.add_revocs = 0;
+          break;
+
+        case oRealClean:
+          opt.realclean = 1;
+          break;
 
 	case aSupported:
 	case aCreate:
@@ -304,10 +324,15 @@ main (int argc, char **argv)
   assuan_set_gpg_err_source (GPG_ERR_SOURCE_DEFAULT);
   setup_libassuan_logging (&opt.debug, NULL);
 
+  opt.add_revocs = 1;  /* Default add revocation certs.  */
+
   /* Parse the command line. */
   pargs.argc  = &argc;
   pargs.argv  = &argv;
   pargs.flags = ARGPARSE_FLAG_KEEP;
+#if GPGRT_VERSION_NUMBER >= 0x013000 /* >= 1.48 */
+  pargs.flags |= ARGPARSE_FLAG_COMMAND;
+#endif
   cmd = parse_arguments (&pargs, opts);
   gpgrt_argparse (NULL, &pargs, NULL);
 
@@ -343,7 +368,7 @@ main (int argc, char **argv)
 
   /* Set defaults for non given options.  */
   if (!opt.gpg_program)
-    opt.gpg_program = gnupg_module_name (GNUPG_MODULE_NAME_GPG);
+    opt.gpg_program = xstrdup (gnupg_module_name (GNUPG_MODULE_NAME_GPG));
 
   if (!opt.directory)
     opt.directory = "openpgpkey";
@@ -353,6 +378,20 @@ main (int argc, char **argv)
                        !opt.no_autostart);
 
 
+  /* Get the short form of the current locale.  */
+  {
+    const char *locname = gnupg_messages_locale_name ();
+    char *p;
+
+    if (locname && *locname && strcmp (locname, "C"))
+      {
+        short_locale = xstrdup (locname);
+        if ((p = strpbrk (short_locale, "_.@/")))
+          *p = 0;
+        gpgrt_annotate_leaked_object (short_locale);
+      }
+  }
+
   /* Check that the top directory exists.  */
   if (cmd == aInstallKey || cmd == aRemoveKey || cmd == aMirror)
     {
@@ -387,7 +426,7 @@ main (int argc, char **argv)
       else
         {
           if (argc != 1)
-            wrong_args ("--supported DOMAIN");
+            wrong_args ("--supported", "DOMAIN");
           err = command_supported (argv[0]);
           if (err && gpg_err_code (err) != GPG_ERR_FALSE)
             log_error ("checking support failed: %s\n", gpg_strerror (err));
@@ -396,15 +435,15 @@ main (int argc, char **argv)
 
     case aCreate:
       if (argc != 2)
-        wrong_args ("--create FINGERPRINT USER-ID");
-      err = command_send (argv[0], argv[1]);
+        wrong_args ("--create", "FINGERPRINT USER-ID");
+      err = command_create (argv[0], argv[1]);
       if (err)
         log_error ("creating request failed: %s\n", gpg_strerror (err));
       break;
 
     case aReceive:
       if (argc)
-        wrong_args ("--receive < MIME-DATA");
+        wrong_args ("--receive", "< MIME-DATA");
       err = wks_receive (es_stdin, command_receive_cb, NULL);
       if (err)
         log_error ("processing mail failed: %s\n", gpg_strerror (err));
@@ -412,7 +451,7 @@ main (int argc, char **argv)
 
     case aRead:
       if (argc)
-        wrong_args ("--read < WKS-DATA");
+        wrong_args ("--read", "< WKS-DATA");
       err = read_confirmation_request (es_stdin);
       if (err)
         log_error ("processing mail failed: %s\n", gpg_strerror (err));
@@ -420,7 +459,7 @@ main (int argc, char **argv)
 
     case aCheck:
       if (argc != 1)
-        wrong_args ("--check USER-ID");
+        wrong_args ("--check", "USER-ID");
       err = command_check (argv[0]);
       break;
 
@@ -437,12 +476,12 @@ main (int argc, char **argv)
       else if (argc == 2)
         err = wks_cmd_install_key (*argv, argv[1]);
       else
-        wrong_args ("--install-key [FILE|FINGERPRINT USER-ID]");
+        wrong_args ("--install-key", "[FILE|FINGERPRINT USER-ID]");
       break;
 
     case aRemoveKey:
       if (argc != 1)
-        wrong_args ("--remove-key USER-ID");
+        wrong_args ("--remove-key", "USER-ID");
       err = wks_cmd_remove_key (*argv);
       break;
 
@@ -1126,6 +1165,19 @@ command_check (char *userid)
                  addrspec);
       err = gpg_error (GPG_ERR_CERT_REVOKED);
     }
+  else if (opt.output)
+    {
+      /* Save to file.  */
+      const char *fname = opt.output;
+
+      if (*fname == '-' && !fname[1])
+        fname = NULL;
+      es_rewind (key);
+      err = wks_write_to_file (key, fname);
+      if (err)
+        log_error ("writing key to '%s' failed: %s\n",
+                   fname? fname : "[stdout]", gpg_strerror (err));
+    }
 
  leave:
   xfree (fpr);
@@ -1140,7 +1192,7 @@ command_check (char *userid)
 /* Locate the key by fingerprint and userid and send a publication
  * request.  */
 static gpg_error_t
-command_send (const char *fingerprint, const char *userid)
+command_create (const char *fingerprint, const char *userid)
 {
   gpg_error_t err;
   KEYDB_SEARCH_DESC desc;
@@ -1304,25 +1356,6 @@ command_send (const char *fingerprint, const char *userid)
         }
     }
 
-
-  /* Now put the armor around the key.  */
-  {
-    estream_t newkey;
-
-    es_rewind (key);
-    err = wks_armor_key (&newkey, key,
-                         no_encrypt? NULL
-                         /* */    : ("Content-Type: application/pgp-keys\n"
-                                     "\n"));
-    if (err)
-      {
-        log_error ("error armoring key: %s\n", gpg_strerror (err));
-        goto leave;
-      }
-    es_fclose (key);
-    key = newkey;
-  }
-
   /* Hack to support posteo but let them disable this by setting the
    * new policy-version flag.  */
   if (policy->protocol_version < 3
@@ -1333,6 +1366,31 @@ command_send (const char *fingerprint, const char *userid)
       posteo_hack = 1;
     }
 
+
+  /* Now put the armor around the key.  */
+  {
+    estream_t newkey;
+    char *prefix;
+
+    prefix = xstrconcat
+      ("Content-Type: application/pgp-keys\n",
+       short_locale && *short_locale? "Content-Language: " : "",
+       short_locale && *short_locale? short_locale : "",
+       short_locale && *short_locale? "\n" : "",
+       "\n", NULL);
+
+    es_rewind (key);
+    err = wks_armor_key (&newkey, key, no_encrypt? NULL : prefix);
+    xfree (prefix);
+    if (err)
+      {
+        log_error ("error armoring key: %s\n", gpg_strerror (err));
+        goto leave;
+      }
+    es_fclose (key);
+    key = newkey;
+  }
+
   /* Encrypt the key part.  */
   if (!no_encrypt)
     {
@@ -1381,8 +1439,10 @@ command_send (const char *fingerprint, const char *userid)
             goto leave;
         }
 
-      err = mime_maker_add_header (mime, "Content-type",
+      err = mime_maker_add_header (mime, "Content-Type",
                                    "application/pgp-keys");
+      if (!err && short_locale && *short_locale)
+        err = mime_maker_add_header (mime, "Content-Language", short_locale);
       if (err)
         goto leave;
 
@@ -1562,10 +1622,11 @@ send_confirmation_response (const char *sender, const char *address,
    * only our client will see it.  */
   if (encrypt)
     {
-      es_fputs ("Content-Type: application/vnd.gnupg.wks\n"
-                "Content-Transfer-Encoding: 8bit\n"
-                "\n",
-                body);
+      es_fprintf (body,
+                  "Content-Type: application/vnd.gnupg.wks\n"
+                  "Content-Transfer-Encoding: 8bit\n"
+                  "Content-Language: %s\n"
+                  "\n", (short_locale && *short_locale)? short_locale : "en");
     }
 
   es_fprintf (body, ("type: confirmation-response\n"
@@ -1759,6 +1820,8 @@ process_confirmation_request (estream_t msg, const char *mainfpr)
       log_info ("no encryption key found - sending response in the clear\n");
       err = send_confirmation_response (sender, address, nonce, 0, NULL);
     }
+  if (!err)
+    log_info ("response sent to '%s' for '%s'\n", sender, address);
 
  leave:
   nvc_release (nvc);
@@ -1822,12 +1885,13 @@ read_confirmation_request (estream_t msg)
 
 /* Called from the MIME receiver to process the plain text data in MSG.  */
 static gpg_error_t
-command_receive_cb (void *opaque, const char *mediatype,
+command_receive_cb (void *opaque, const char *mediatype, const char *language,
                     estream_t msg, unsigned int flags)
 {
   gpg_error_t err;
 
   (void)opaque;
+  (void)language;
   (void)flags;
 
   if (!strcmp (mediatype, "application/vnd.gnupg.wks"))
@@ -1883,7 +1947,7 @@ domain_matches_mbox (const char *domain, const char *mbox)
  * so that for a key with
  *    uid: Joe Someone <joe@example.org>
  *    uid: Joe <joe@example.org>
- * only the news user id (and thus its self-signature) is used.
+ * only the newest user id (and thus its self-signature) is used.
  * UIDLIST is nodified to set all MBOX fields to NULL for a processed
  * user id.  FPR is the fingerprint of the key.
  */
@@ -1990,7 +2054,7 @@ mirror_one_key (estream_t key)
         continue; /* No mail box or already processed.  */
       if (uid->expired)
         continue;
-      if (!domain_matches_mbox (domain, uid->mbox))
+      if (*domain && !domain_matches_mbox (domain, uid->mbox))
         continue; /* We don't want this one.  */
       if (is_in_blacklist (uid->mbox))
         continue;
